# -*- coding: utf-8 -*-
"""
Routines for parsing LEMI field spreadsheets(i.e. install and demobilization
sheets) and validating metadata;
Templates for the spreadsheets are provided to the PIs and their formatting is
hard coded.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

import copy
import openpyxl

from dataclasses import fields
from datetime import datetime
from obspy import UTCDateTime
from openpyxl.worksheet.worksheet import Worksheet
from pathlib import Path
from typing import Dict, List, Optional, TYPE_CHECKING, Union

from lemi2seed.logging import parse_config_ini, setup_logger
from lemi2seed.metadata_category import Aux, Elec, Mag, Run, Sta, Net
from lemi2seed.utils import is_empty, get_e_ids, get_run_list, NUM_E_CHA_MAX

if TYPE_CHECKING:
    import numpy
    from lemi2seed.lemi_data import LemiData

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)

# -- Type aliases --
MD_VAL = Union[str, int, float, datetime, UTCDateTime]  # Type of metadata value
DCS = Union[Net, Sta, Run, Elec, Mag, Aux]  # Type of dataclasses
DCS_SUB_REMA = Union[Run, Elec, Mag, Aux]
DCS_SUB_EMA = Union[Elec, Mag, Aux]
DCS_SUB_MA = Union[Mag, Aux]
# -- Type aliases --

CHA_TYPES = {'elec': ['Ex', 'Ey'], 'mag': ['Hx', 'Hy', 'Hz'],
             'aux': ['Ui', 'Te', 'Tf', 'Sn', 'Fq', 'Ce']}


class LemiMetadata():
    """
    This class handles the parsing the LEMI field sheets, the population of the
    metadata and their validation.
    """

    def __init__(self, path2md: Path, output_xml: Path, lemi_data: LemiData) -> None:
        """Instantiate LemiMetadata class and scan metadata directory."""
        self.path2md = path2md
        self.output_xml = output_xml
        self.filenames: List[Path] = []
        self.cats: List[str] = ['Net', 'Sta', 'Run', 'Elec', 'Mag', 'Aux']
        self.net: Net = Net()
        self.sta: Sta = Sta()
        self.run: List[Run] = []
        self.elec: List[Elec] = []
        self.mag: List[Mag] = []
        self.aux: List[Aux] = []
        self.data_stats: Dict = lemi_data.stats
        self.data_np: numpy.ndarray = lemi_data.data_np
        self.run_list: List[str] = get_run_list(lemi_data.stats['num_runs'])
        if path2md is not None:
            self.scan_path2md()

    def scan_path2md(self) -> None:
        """
        Scan metadata directory. If no field sheets are found under the given
        directory, the user will have to provide the necessary metadata in the
        GUI.
        """
        filenames = [tmp for tmp in self.path2md.glob('*.xlsx')
                     if not tmp.parts[-1].startswith('~$')]
        if not filenames:
            logger.error("No field sheet found under the following path - {}. "
                         "All metadata will have to be provided using the "
                         "GUI!".format(self.path2md))
        else:
            self.filenames = filenames

    @staticmethod
    def get_md_field(sheet: Worksheet, cell: str) -> Optional[MD_VAL]:
        """Get metadata field at a given cell in a given field sheet."""
        try:
            field_val = sheet[cell].value
        except Exception as error:
            logger.exception("Failed to retrieve value of cell {} - Skipping "
                             "cell - Exception: {}".format(cell, error))
            return None
        return field_val

    def identify_field_sheet(self, sheet: Worksheet, filename: Path) -> Optional[str]:
        """
        Check whether a field sheet is an install, removal or credential sheet.
        Layout between provided spread sheet templates varies and cell numbers
        listed in config file are organized in three sub-categories:
        install_sheet, credential_sheet, removal_sheet.
        """
        sheet_header = LemiMetadata.get_md_field(sheet, config['sheets']['header'])
        sheet_types = {val: key for key, val in config['sheets'].items()[1:]}
        sheet_type = sheet_types.get(sheet_header)
        if sheet_type is None:
            logger.warning("The following file {} does not have the proper "
                           "header. The provided spread sheet templates were "
                           "not used or their layout was modified. Skipping "
                           "file!".format(filename))
        return sheet_type

    def init_run_md_props(self) -> List[Run]:
        """
        Instantiate Run data class object and initialize non-user defined
        properties.
        """
        runs = []
        run_num = [x['run_num'] for x in self.data_np]
        starttime = [x['starttime'] for x in self.data_np]
        endtime = [x['endtime'] for x in self.data_np]
        for ind_run, run_id in enumerate(self.run_list, 1):
            run = Run(resource_id=f'mt.run.id:{run_id}',
                      start=starttime[run_num.index(ind_run)],
                      end=endtime[len(run_num) - run_num[::-1].index(ind_run) - 1])
            runs.append(run)
        return runs

    def init_cha_md_props(self, cha_type: str, run_id: str) -> DCS_SUB_EMA:
        """
        Create instances of Elec, Mag or Aux data class and
        initialize properties inherited from the BaseSta and BaseChannel
        data classes.
        """
        ChaClass = globals()[cha_type.capitalize()]
        cha = ChaClass(elev=round(self.data_stats["elev"], 3),
                       lat=round(self.data_stats["lat"], 3),
                       lon=round(self.data_stats["lon"], 3),
                       run_id=run_id)
        return cha

    @staticmethod
    def dc2dict(dc: DCS_SUB_REMA) -> Dict:
        """
        Convert dataclass instance to a dict and filter dict to:
        - only keep metadata fields displayed in GUI
        - reset missing and invalid metadata fields
        """
        missing_invalid = dc.md_missing | dc.md_invalid
        mdict = {}
        for md_field in fields(dc):
            if md_field.metadata.get('gui'):
                if md_field.name not in missing_invalid:
                    mdict[md_field.name] = getattr(dc, md_field.name)
                else:
                    mdict[md_field.name] = md_field.default
        return mdict

    def filter_cha(self, cha_type: str, run_id: str) -> List[DCS_SUB_EMA]:
        """
        For a given channel type (electric, magnetic or auxiliary), get list of
        channels with a given run id.
        """
        return [c for c in getattr(self, cha_type) if c.run_id == run_id]

    @staticmethod
    def update_cha_num(cha: Elec, cha_num: str) -> Elec:
        """
        Make copy of populated Elec data classes and set channel number.
        """
        tmp = copy.deepcopy(cha)
        tmp.cha_num = cha_num
        return tmp

    @property
    def for_gui(self) -> Dict:
        """
        Define metadata output dictionary structure that will be used to set
        the model in the GUI.
        """
        output: Dict = {}
        for cat in self.cats[:-1]:
            output[cat] = {}
            mprop = getattr(self, cat.lower())
            if isinstance(mprop, list):
                if cat == 'Run':
                    if not mprop:
                        mprop = self.init_run_md_props()
                    for ind, dc in enumerate(mprop):
                        key = f'Run_{self.run_list[ind]}'
                        output[cat][key] = LemiMetadata.dc2dict(dc)
                elif cat == 'Elec':
                    for run_id in self.run_list:
                        chas = self.filter_cha('elec', run_id)
                        if not chas:
                            cha = self.init_cha_md_props('elec', run_id)
                            chas = [LemiMetadata.update_cha_num(cha, f'E{ind + 1}')  # type: ignore
                                    for ind in range(NUM_E_CHA_MAX)]  # type: ignore
                        for ind in range(NUM_E_CHA_MAX):
                            key = f'Run_{run_id}_Elec_Pair_{ind + 1}'
                            output[cat][key] = LemiMetadata.dc2dict(chas[ind])
                elif cat == 'Mag':
                    for run_id in self.run_list:
                        chas = self.filter_cha('mag', run_id)
                        if not chas:
                            dc = self.init_cha_md_props('mag', run_id)
                        else:
                            dc = chas[0]
                        key = f'Run_{run_id}_Mag'
                        output[cat][key] = LemiMetadata.dc2dict(dc)
            else:
                output[cat] = LemiMetadata.dc2dict(mprop)
        return output

    def parse_field_sheet(self, sheet: Worksheet, sheet_type: str, md_fields: Dict) -> Dict:
        """
        Parse field sheet using the cell numbers listed in the config file and
        return a nested dictionary of metadata entries organized based on the
        category under which a given metadata field falls.
        """
        logger.info("Parsing the {} {}".format(*sheet_type.split('_')))
        for cat in self.cats[:-1]:
            if cat not in md_fields:
                md_fields[cat] = {}
            if sheet_type not in config[cat]:
                continue
            logger.info("Parsing '{}' metadata".format(cat))
            for key, val in config[cat][sheet_type].items():
                if isinstance(val, list):
                    field_val = [LemiMetadata.get_md_field(sheet, x) for x in val]
                else:
                    field_val = LemiMetadata.get_md_field(sheet, val)  # type: ignore
                if key == "comments":
                    field_val = ', '.join(filter(None, (md_fields[cat].get(key, None), field_val)))  # type: ignore
                md_fields[cat][key] = field_val
        return md_fields

    def reformat_run(self, run_fields: Dict) -> Dict:
        """Reformat run metadata fields."""
        tmp = {f'Run_{run_id}': dict.fromkeys(run_fields.keys()) for run_id in self.run_list}
        tmp['Run_a'].update(run_fields)
        return tmp

    def reformat_elec(self, elec_fields: Dict) -> Dict:
        """Reformat electric metadata fields."""
        e_ids = get_e_ids(self.data_stats['num_runs'])
        tmp = {f'Run_{run_id}_Elec_Pair_{ind}': dict.fromkeys(elec_fields.keys())
               for run_id, ind in e_ids}
        for key, val in elec_fields.items():
            for ind in range(1, NUM_E_CHA_MAX + 1):
                if isinstance(val, list):
                    e_val = val[ind - 1]
                else:
                    e_val = val
                tmp[f'Run_a_Elec_Pair_{ind}'][key] = e_val
        return tmp

    def reformat_mag(self, mag_fields: Dict) -> Dict:
        """Reformat magnetic metadata fields."""
        tmp = {f'Run_{run_id}_Mag': dict.fromkeys(mag_fields.keys())
               for run_id in self.run_list}
        tmp['Run_a_Mag'].update(mag_fields)
        return tmp

    def reformat_md_dict(self, md_fields: Dict) -> Dict:
        """
        Reformat metadata_fields to include all runs and associate correct
        metadata info to each electrode pair.
        """
        md_fields['Run'] = self.reformat_run(md_fields['Run'])
        md_fields['Elec'] = self.reformat_elec(md_fields['Elec'])
        md_fields['Mag'] = self.reformat_mag(md_fields['Mag'])
        return md_fields

    @property
    def from_field_sheets(self) -> Optional[Dict]:
        """Parse all field sheets provided by the user."""
        md_fields: Dict = {}
        parsed_sheets = []
        for filename in self.filenames:
            try:
                workbook = openpyxl.load_workbook(filename, data_only=True)  # type: ignore
            except Exception as error:
                logger.error("Failed to open field sheet: {} - Skipping file "
                             "- Exception: {}".format(filename, error))
            else:
                if workbook is None:
                    continue
                sheet = workbook.active
                sheet_type = self.identify_field_sheet(sheet, filename)
                if sheet_type is None:
                    continue
                if sheet_type not in parsed_sheets:
                    parsed_sheets.append(sheet_type)
                else:
                    logger.warning("Already parsed the {0} {1} - You may have "
                                   "more than one {0} {1} - Skipping {2}!"
                                   .format(*sheet_type.split('_'), filename))
                    continue
                md_fields = self.parse_field_sheet(sheet, sheet_type, md_fields)
                workbook.close()
        return self.reformat_md_dict(md_fields) if not is_empty(md_fields) else None
